# encoding: utf-8

module Majic
  module Logging
    Levels      = [:debug, :info, :warn, :error, :fatal]
    LevelColors = { debug: :white, info: :blue, warn: :yellow, error: :red, fatal: :orange }

    DefaultNameCutoff = 18
    UnknownLevelError = Class.new RuntimeError

    @level  = Levels.first
    @stream = $stdout
    @mutex  = Mutex.new

    class << self
      def included klass
        klass.metaclass.send :attr_accessor, :logger
        klass.logger = Logger.new klass
      end

      attr_accessor :level, :stream
      attr_reader :mutex

      attr_writer :name_cutoff
      def name_cutoff; @name_cutoff or DefaultNameCutoff end
      def name_cutoff?; !!@name_cutoff end

      def level= level
        raise UnknownLevelError unless Levels.include? level
        @level = level
      end
      
      def level_index
        Levels.index level
      end
    end
    
    def log *messages
      logger_instance = (metaclass.respond_to?(:logger) && metaclass.logger or self.class.logger)
      
      if messages.empty?
        logger_instance
      else
        logger_instance.debug *messages
      end
    end
    
    class Logger
      TimeFormat = '%H:%M:%S'
      
      attr_writer :name
      
      @@names = Set.new
      
      class << self
        def name_width; Logging.name_cutoff or @@names.map(&:length).max end
      end
      
      def initialize object
        @object = object
        @@names << name
      end
      
      def log level, *messages
        return if not should_message? level
        
        Logging.mutex.synchronize do
          messages.each do |message|
            Logging.stream.puts format(level, message).force_encoding 'utf-8'
          end
        end
      end
      
      Levels.each do |level|
        define_method level do |*messages|
          log level, *messages
        end
      end
      
      def name_self object, name
        # FIXME: This appears not to be called when attaching to a metaclass.
        #        An investigation is in order.
        Logging.included object.metaclass
        
        object.extend Logging
        object.log.name = name
      end
      
      def name
        @name or default_name
      end
      
      def name= name
        @@names.delete @name
        @name = name
        @@names << @name
      end
      
    protected
      
      def default_name
        cutoff = Logging.name_cutoff
        
        default_name = if Module === @object
            name = @object.to_s
            name.sub! /^.+?::/, '' while name.include? '::' and name.length > cutoff
            name
          else
            @object.to_s
          end
        
        default_name.truncate! cutoff, :middle if default_name.length > cutoff
        
        default_name
      end
      
      def format level, message
        name = self.name.ljust(Logger.name_width) ^ :bold
        [timestamp, name, level_string(level), " #{message}"].join " "
      end
      
      def should_message? level
        Levels.index(level) >= Logging.level_index
      end
      
      def timestamp
        Time.now.strftime TimeFormat
      end
      
      def level_string level
        max_width = Levels.map(&:length).max
        level.to_s.capitalize.rjust(max_width) ^ LevelColors[level]
      end
    end
  end
end
